/*********************************************************************
 * This file is part of the cpplibs suite.
 *
 * Copyright (C) 2001 Topi Menp
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.3 $
 *********************************************************************/

#ifndef _BLOB_H
#define _BLOB_H

#include "Object.h"

namespace util
{
	/**
	 * Blob (Binary Large OBject) is a class for storing a pointer to
	 * data and the size of the data. As a template, Blob can store
	 * pointers to any type of data.
	 **/
	template <class T> class Blob : virtual public Object
	{
	public:
		/**
		 * Create a new Blob.
		 *
		 * @param ptr a pointer to data
		 * @param length the number of items that can be sequentially accessed at the pointer
		 * @param release if true, the memory pointed to by <i>ptr</i> is released upon
		 * deletion of the Blob
		 **/
		Blob(T* ptr = NULL, int len = 0, bool release = false) : _pData(ptr), _iLen(len), _bRelease(release) {}
		/**
		 * Copy a Blob. The memory pointed by the pointer in <i>other</i>
		 * is not copied. Instead, the two Blobs now reference the same
		 * memory location.
		 **/
		Blob(const Blob& other) : _pData(other._pData), _iLen(other._iLen), _bRelease(false) {}

		/**
		 * Destroy a Blob.
		 **/
		~Blob() { if (_bRelease) delete _pData; }

		/**
		 * Copy a blob. If the release flag is true, the memory pointer is
		 * deleted.
		 **/
		Blob& operator= (const Blob& other)
		{
			if (other._pData != _pData)
				{
					if (_bRelease) delete _pData;
					_pData = other._pData;
					_iLen = other._iLen;
					_bRelease = false;
				}
		}

		/**
		 * Get a pointer to the internal data.
		 **/
		const T* getData() const { return _pData; }
		/**
		 * Get a pointer to the internal data.
		 **/
		T* getData() { return _pData; }
		/**
		 * Get the size of the data.
		 **/
		int getLength() const { return _iLen; }
		/**
		 * Release the data. The release flag is set to false, and the
		 * internal pointer is returned.
		 **/
		T* release() { _bRelease = false; return _pData; }

	private:
		T* _pData;
		int _iLen;
		bool _bRelease;
	};
}

#endif
