/*********************************************************************
 * This file is part of the cpplibs suite.
 *
 * Copyright (C) 2001 Topi Menp
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.6 $
 *********************************************************************/

#include "Expression.h"

#include <StreamTokenizer.h>
#include <Number.h>
#include <sstream>

using namespace std;
using namespace util::io;

namespace util { namespace ariel {

	bool Expression::_bInitialized = false;
	StreamTokenizer Expression::_tokenizer;
	VariableManager* Expression::_pVariableManager = NULL;
	Pair<int,BinaryFunction*> Expression::_opFunctionCall = Pair<int,BinaryFunction*>(7,new FunctionCall);
	Pair<int,BinaryFunction*> Expression::_opArrayIndex = Pair<int,BinaryFunction*>(7,new ArrayIndex);
	Pair<int,BinaryFunction*> Expression::_opNonFirstComma = Pair<int,BinaryFunction*>(0,new Comma(false));
	UnaryFunction* Expression::_opNegate = new Negate;
	UnaryFunction* Expression::_opComplement = new Complement;
	UnaryFunction* Expression::_opFetchValue = new FetchValue;
	Hashtable<string,Computable*> Expression::_tblInternalVariables;

	class OperatorSection : public Section
	{
	public:
		OperatorSection(string start, Pair<int,BinaryFunction*> op) : Section("operator",start), operation(op) {}
		string readSection(istream& input) throw(IOException&) { return ""; }

		Pair<int,BinaryFunction*> operation;
	};
	

	void Expression::createSection(string start, BinaryFunction* op, int index)
	{
		_tokenizer.addSection(new OperatorSection(start,Pair<int,BinaryFunction*>(index,op)));
	}

	void Expression::createFunction(string name, Function::MathFunction::FPointer func)
	{
		_tblInternalVariables.put(name, new Function::MathFunction(func,name));
	}

	void Expression::init()
	{
		if (!_bInitialized)
			{
				_bInitialized = true;

				createSection(",",new Comma,0);

				createSection("=",new Assignment,1);

				createSection("&",new BinaryAnd,2);
				createSection("|",new BinaryOr,2);
				createSection("^",new BinaryXor,2);

				createSection("&&",new LogicalAnd,3);
				createSection("||",new LogicalOr,3);

				createSection("==",new Equal,4);
				createSection("!=",new Different,4);

				createSection("<",new Less,5);
				createSection(">",new Greater,5);
				createSection("<=",new LessOrEqual,5);
				createSection(">=",new GreaterOrEqual,5);

				createSection("+",new Add,6);
				createSection("-",new Subtract,6);

				createSection("*",new Multiply,7);
				createSection("/",new Divide,7);
				createSection("%",new Modulus,7);

				createFunction("abs",fabs);
				createFunction("log",log);
				createFunction("exp",exp);
				createFunction("sqrt",sqrt);
				createFunction("sin",sin);
				createFunction("cos",cos);
				createFunction("tan",tan);
				createFunction("asin",asin);
				createFunction("acos",acos);
				createFunction("atan",atan);

				_tblInternalVariables.put("false",new Boolean(false));
				_tblInternalVariables.put("true",new Boolean(true));
				_tblInternalVariables.put("void",new Void);
				
				_tblInternalVariables.put("pi",new Number::Double(M_PI));
			}
	}


	Expression::Expression()
	{
		init();
	}

	Expression::Expression(string expression) : _strExpression(expression)
	{
		init();
	}

	Expression::Expression(SmartPtr<Computable> value, List<UnaryFunction*> funcs) throw (ComputationException&) :
		_currentValue(value), _operator(NULL), _left(NULL), _right(NULL), _lstFunctions(funcs)
	{
		if (value.get())
			for (int i=_lstFunctions.getLength();i--;)
				_currentValue = (*_lstFunctions[i])(_currentValue);
	}

	Computable& Expression::parse(string expression) throw (ParseException&,ComputationException&)
	{
		_left.reset();
		_right.reset();
		_strExpression = expression;
		istringstream in(_strExpression);
		readExpression(in);

		return evaluate();
	}

	Pair<int,BinaryFunction*> Expression::getOperator(StreamTokenizer::Token* token)
	{
		if (token->getType() == TOKEN_SECTION)
			{
				StreamTokenizer::SectionToken* section = (StreamTokenizer::SectionToken*)token;
				if (section->getName() == "operator")
					{
						if (section->getSection()->getStart() == "," && _iCommaCnt++) //comma needs special attention
							return _opNonFirstComma;
						return ((OperatorSection*)(section->getSection()))->operation;
					}
			}
		return Pair<int,BinaryFunction*>(0,NULL);
	}

	UnaryFunction* Expression::getUnaryFunction(char chr)
	{
		switch (chr)
			{
			case '-': return _opNegate;
			case '!': return _opComplement;
			case '$': return _opFetchValue;
			}
		return NULL;
	}


	//Read a part of an expression from 'in' until 'end' is reached. 0 =
	//eof.
	void Expression::readExpression(istream& in, char end) throw (ParseException&)
	{
		//An expression consists of a sequence of 'atoms' that are either
		//'items' or 'operators'. The parser switches between these possibilities
		enum State { READING_ITEM, READING_OPERATOR };
		//First atom must be an item.
		State state = READING_ITEM;

		SmartPtr<StreamTokenizer::Token> token;

		//A list for unary operators that is fed to the next item.
		List<UnaryFunction*> unaryList(4,4);

		//Lists for items and operators
		List<Expression*> itemList(32,32);
		List<Pair<int,BinaryFunction*> > operatorList(32,32); //stores operator and its precedence

		_iCommaCnt = 0; //reset comma counter
		try
			{
				while (true)
					{
						//Read next token
						char chr;
						UnaryFunction* op;
						StreamTokenizer::SectionToken* section;
						token = _tokenizer.getNextToken(in);
						if (state == READING_ITEM) //We are expecting an item
							{
								switch (token->getType())
									{
									case TOKEN_WORD: //word
										{
											//cerr << "Word " << ((StreamTokenizer::WordToken*)token.get())->getWord() << endl;
											string word = ((StreamTokenizer::WordToken*)token.get())->getWord();
											Computable** variable = _tblInternalVariables.get(word);
											if (variable)
												itemList += new Expression(SmartPtr<Computable>(*variable,false),unaryList);
											else
												itemList += new Expression(SmartPtr<Computable>(new Variable(word)),
																									 unaryList);
											unaryList.clear();
											state = READING_OPERATOR; //switch to reading an operator
										}
										break;
								
									case TOKEN_NUMBER: //number
										//cerr << "Number " << ((StreamTokenizer::NumberToken*)token.get())->getNumber() << endl;
										itemList += new Expression(SmartPtr<Computable>(new Number::Double(((StreamTokenizer::NumberToken*)token.get())->getNumber())),
																							 unaryList);
										unaryList.clear();
										state = READING_OPERATOR; //switch to reading an operator
										break;
								
									case TOKEN_NORMAL: //opening bracket
										chr = ((StreamTokenizer::NormalToken*)token.get())->getCharacter();
										op = getUnaryFunction(chr);
										if (op)
											{
												unaryList += op;
												break;
											}
										else if (chr == '(')
											{
												SmartPtr<Expression> expr(new Expression(SmartPtr<Computable>(NULL),unaryList));
												unaryList.clear();
												expr->readExpression(in,')');
												itemList += expr.release();
											}
										else if (chr == end && !itemList.getLength()) //end of an empty expression
											{
												itemList += new Expression(SmartPtr<Computable>(new Void),unaryList);
												goto out;
											}
										else
											throw ParseException(string("Expression::readExpression(istream&,char): Invalid character: ") + chr);
										state = READING_OPERATOR; //switch to reading an operator
										break;
								
									case TOKEN_SECTION:
										section = (StreamTokenizer::SectionToken*)token.get();
										if (section->getName() == "string") //string constant
											{
												itemList += new Expression(SmartPtr<Computable>(new String(section->getContents())),unaryList);
												unaryList.clear();
												state = READING_OPERATOR;
											}
										else //unary operator (minus)
											{
												//cerr << section->getSection()->getStart()[0] << endl;
												op = getUnaryFunction(section->getSection()->getStart()[0]);
												if (op)
													unaryList += op;
												else
													throw ParseException("Expression::readExpression(istream&,char): Found '" +
																							 section->getSection()->getStart() + section->getContents() +
																							 "' when expecting a number, bracket, string, function call or unary operator.");
											}
										break;
									}
							} //READING_ITEM
						else if (state == READING_OPERATOR) //We are expecting an operator
							{
								Pair<int,BinaryFunction*> op;
								StreamTokenizer::SectionToken* section;
								switch (token->getType())
									{
									case TOKEN_SECTION: //binary operator
										section = (StreamTokenizer::SectionToken*)token.get();
										op = getOperator(token);
										if (!op.second())
											throw ParseException("Expression::readExpression(istream&,char): Found '" +
																					 section->getSection()->getStart() + section->getContents() +
																					 "' when expecting a binary operator.");
										operatorList += op;
										state = READING_ITEM; //switch back to reading an item
										break;

									case TOKEN_NORMAL:
										chr = ((StreamTokenizer::NormalToken*)token.get())->getCharacter();
										if (chr == end) //done reading
											goto out;
										//opening bracket (operator() or operator[])
										//bracketed expressions are treated as ordinary operators
										else if (chr == '(' || chr == '[') 
											{
												SmartPtr<Expression> expr(new Expression());
												expr->readExpression(in,chr == '(' ? ')' : ']');
												operatorList += (chr == '(' ? _opFunctionCall : _opArrayIndex);
												itemList += expr.release();
											}
										else
											throw ParseException(string("Expression::readExpression(istream&,char): Illegal character: ") +
																					 chr + " when excepting a binary operator.");
										break;
										
									case TOKEN_WORD:
										throw ParseException("Expression::readExpression(istream&,char): Missing an operator before " +
																				 ((StreamTokenizer::WordToken*)token.get())->getWord());
									case TOKEN_NUMBER:
										throw ParseException("Expression::readExpression(istream&,char): Missing an operator before " +
																				 ((StreamTokenizer::NumberToken*)token.get())->getNumberString());
									}
							} //READING_OPERATOR
					} //while tokens
			out:;
			}
		catch (EOFException& eof)
			{
				if (state != READING_OPERATOR || end) //cannot end after operator or when reading an subexpression
					throw ParseException ("Expression::readExpression(istream&,char): Unexpected end of input.");
			}
		catch (ComputationException& ce)
			{
				throw ParseException(ce.getMessage());
			}
		catch (IOException& ioe)
			{
				throw ParseException("Expression::readExpression(istream&,char): parse error: " + ioe.getMessage());
			}

		if (operatorList.getLength())
			split(itemList,operatorList);
		else if (itemList.getLength() == 1)
			_left = itemList[0];
	}


	void Expression::split(List<Expression*> items, List<Pair<int,BinaryFunction*> > operators)
	{
		//cerr << "Splitting " << items.getLength() << " items separated with " << operators.getLength() << " operators." << endl;
		int firstIndex = -1;

		//Find the operator to be applied _last_. The expression is first
		//split into two parts on the operator that must be applied last.
		int minRank = 100;
		for (int i=operators.getLength();i--;)
			{
				int rank = operators[i].first();
				if (rank < minRank ||
						(rank <= minRank && operators[i].second()->executionOrder == BinaryFunction::ORDER_RIGHT_TO_LEFT))
					{
						firstIndex = i;
						minRank = rank;
					}
			}

		//cerr << "Splitting at " << (firstIndex+1) << endl;

		if (firstIndex >= 0)
			_operator = (BinaryFunction*)operators[firstIndex].second();
		
		if (firstIndex > 0)
			{
				_left = new Expression;
				_left->split(items(0,firstIndex+1),operators(0,firstIndex));
			}
		else
			_left = items[0];

		if (firstIndex < operators.getLength()-1)
			{
				_right = new Expression;
				_right->split(items(firstIndex+1),operators(firstIndex+1));
			}
		else
			_right = items[firstIndex+1];
	}
	
	Computable& Expression::evaluate() throw (ComputationException&)
	{
		if (!_left && !_right)
			return _currentValue;
		
		if (_left && _right && _operator)
			{
				_currentValue = (*_operator)(_left->evaluate(),
																		 _right->evaluate());
			}
		else if (_left)
			{
				_left->evaluate();
				_currentValue = _left->_currentValue.get();
				_currentValue.release();
			}

		for (int i=_lstFunctions.getLength();i--;)
			_currentValue = (*_lstFunctions[i])(_currentValue);

		return _currentValue;
	}

	SmartPtr<Computable> Variable::setValue(const Computable& value) const throw (ComputationException&)
	{
		if (!Expression::_pVariableManager)
			throw ComputationException("Variable::setValue(const Computable&): There is no variable manager. "
																 "Cannot set value for '" + _strName + "'.");
		return Expression::_pVariableManager->setValue(_strName,value);
	}

	SmartPtr<Computable> Variable::getValue() const throw (ComputationException&)
	{
		if (!Expression::_pVariableManager)
			throw ComputationException("Variable::getValue(): There is no variable manager. Cannot get a value for '" + _strName + "'.");
		return Expression::_pVariableManager->getValue(_strName);
	}

	SmartPtr<Computable> VariableManager::Simple::getValue(std::string name) throw (ComputationException&)
	{
		SmartPtr<Computable>* value = get(name);
		if (!value)
			throw ComputationException("VariableManager::Simple::getValue(string): No value set for '" + name + "'.");
		return SmartPtr<Computable>(value->get(),false);
	}

	SmartPtr<Computable> VariableManager::Simple::setValue(std::string name, const Computable& value) throw (ComputationException&)
	{
		try
			{
				Computable* v = dynamic_cast<Computable*>(value.clone());
				put(name,SmartPtr<Computable>(v));
				return SmartPtr<Computable>(v,false);
			}
		catch(NotCloneableException& nce)
			{
				throw ComputationException(nce.getMessage());
			}
	}
	
}}
