/*********************************************************************
 * This file is part of the cpplibs suite.
 *
 * Copyright (C) 2001 Topi Menp
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.4 $
 *********************************************************************/

#ifndef _OPERATOR_H
#define _OPERATOR_H

#include <SmartPtr.h>
#include <Computable.h>
#include <Boolean.h>
#include <Object.h>

namespace util { namespace ariel {

	/**
	 * Operator is a base class for binary and unary operators.
	 **/
	class Operator : virtual public Object
	{
	public:
	};

	/**
	 * UnaryFunction is an operator that works with a single argument.
	 **/
	class UnaryFunction : public Operator
	{
	public:
		/**
		 * Perform an unary operation.
		 **/
		virtual SmartPtr<Computable> operator() (const Computable& c) throw (ComputationException&) = 0;
	};

	/**
	 * Negate negates its argument.
	 **/
	class Negate : public UnaryFunction
	{
	public:
		SmartPtr<Computable> operator() (const Computable& c) throw (ComputationException&) { return c.negate(); }
	};

	/**
	 * Negate complements its argument.
	 **/
	class Complement : public UnaryFunction
	{
	public:
		SmartPtr<Computable> operator() (const Computable& c) throw (ComputationException&) { return c.complement(); }
	};

	/**
	 * The '$' operator. Fetches a value for a variable.
	 **/
	class FetchValue : public UnaryFunction
	{
	public:
		SmartPtr<Computable> operator() (const Computable& c) throw (ComputationException&);
	};


	/**
	 * BinaryFunction is a base class for operations that have two
	 * operands.
	 **/
	class BinaryFunction : public Operator
	{
	public:
		/**
		 * An enumeration for two possible execution orders. (Note that
		 * 'execution order' does not mean 'operator precedence'.) For
		 * some operations, the execution order plays no significant role.
		 * It makes no difference to evaluate '1+2+3' from either left to
		 * right or right to left. However, operators like '-', ',' and
		 * '=' depend on a strictly defined execution order. '=' must be
		 * evaluated right to left, because otherwise 'a=b=1' would yield
		 * no value for 'a'. At the same time, '(a,(b,c),d)' must be
		 * evaluated from left to right, because otherwise the result
		 * would be '(a,(b,c,d))'. In the most obvious case, '3-2-1' is
		 * different from '3-(2-1)'. Possible values for execution order
		 * are:
		 *
		 * <ul>
		 * <li>ORDER_LEFT_TO_RIGHT</li>
		 * <li>ORDER_RIGHT_TO_LEFT</li>
		 * </ul>
		 **/
		enum ExecutionOrder { ORDER_LEFT_TO_RIGHT, ORDER_RIGHT_TO_LEFT };
		/**
		 * All binary functions perform operations on to operands and must
		 * override this method.
		 **/
		virtual SmartPtr<Computable> operator()(const Computable& c1, const Computable& c2) throw (ComputationException&) = 0;
		/**
		 * The execution order for this function
		 **/
		ExecutionOrder executionOrder;

	protected:
		/**
		 * Create a new binary function that has the specified execution
		 * order.
		 **/
		BinaryFunction(ExecutionOrder order=ORDER_LEFT_TO_RIGHT) : executionOrder(order) {}
	};

	/**
	 * The Comma operator (,) is used to separate expressions in lists.
	 * It works in two modes: 'first' and 'non-first'. The first comma
	 * in an expression always concatenates its left and right operands
	 * into a list. All subsequent commas add their right operand to the
	 * list given as the left operand. For 'non-first' commas, the left
	 * operand must always be a list.
	 **/
	class Comma : public BinaryFunction
	{
	public:
		Comma(bool first=true) : _bFirst(first) {}
		SmartPtr<Computable> operator()(const Computable& c1, const Computable& c2) throw (ComputationException&);

		void setFirst(bool first) { _bFirst = first; }
		bool isFirst() { return _bFirst; }
	private:
		bool _bFirst;
	};

	/**
	 * Add sums two operands together.
	 **/
	class Add : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a+b; }
	};

	/**
	 * Subtract calculates the difference between two operands.
	 **/
	class Subtract : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a-b; }
	};

	/**
	 * Multiply multiplies two operands.
	 **/
	class Multiply : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a*b; }
	};

	/**
	 * Divide calculates the quotient of two operands.
	 **/
	class Divide : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a/b; }
	};

	/**
	 * Divide calculates the remainder of two operands after division.
	 **/
	class Modulus : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a%b; }
	};

	/**
	 * Equal compares two operands for equivalence.
	 **/
	class Equal : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a==b; }
	};

	/**
	 * Different compares two operands for difference.
	 **/
	class Different : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator ()(const Computable& a, const Computable& b) throw (ComputationException&) { return a!=b; }
	};

	/**
	 * Less checks if the first operand is smaller than the second one.
	 **/
	class Less : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a<b; }
	};

	/**
	 * Greater checks if the first operand is greater than the second
	 * one.
	 **/
	class Greater : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a>b; }
	};

	/**
	 * LessOrEqual checks if the first operand is smaller than or equal
	 * to the second one.
	 **/
	class LessOrEqual : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a<=b; }
	};

	/**
	 * GreaterOrEqual checks if the first operand is greater than or
	 * equal to the second one.
	 **/
	class GreaterOrEqual : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a>=b; }
	};

	/**
	 * LogicalAnd performs a logical and operation on two arguments.
	 **/
	class LogicalAnd : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a && b; }
	};

	/**
	 * LogicalOr performs a logical or operation on two arguments.
	 **/
	class LogicalOr : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a || b; }
	};
	
	/**
	 * BinaryAnd performs a binary and operation on two arguments.
	 **/
	class BinaryAnd : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a & b; }
	};

	/**
	 * BinaryOr performs a binary or operation on two arguments.
	 **/
	class BinaryOr : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a | b; }
	};

	/**
	 * BinaryXor performs a binary exclusive or operation on two
	 * arguments.
	 **/
	class BinaryXor : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&) { return a ^ b; }
	};

	/**
	 * FunctioCall calls the first operand with the second one as an
	 * argument.
	 **/
	class FunctionCall : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&)	{	return a(b); }
	};

	/**
	 * ArrayIndex uses the second operand as an array index to the first
	 * one.
	 **/
	class ArrayIndex : public BinaryFunction
	{
	public:
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&)	{ return a[b]; }
	};

	/**
	 * Assignment assigns the second operand to the first one.
	 **/
	class Assignment : public BinaryFunction
	{
	public:
		Assignment() : BinaryFunction(ORDER_RIGHT_TO_LEFT) {}
		SmartPtr<Computable> operator()(const Computable& a, const Computable& b) throw (ComputationException&);
 	};
}}

#endif
